# @Author:  Felix Kramer
# @Date:   2021-06-03T11:02:57+02:00
# @Email:  kramer@mpi-cbg.de
# @Project: go-with-the-flow
# @Last modified by:    Felix Kramer
# @Last modified time: 2021-08-30T23:38:31+02:00
# @License: MIT

import sys
import networkx as nx
import numpy as np
import scipy.linalg as lina
import random as rd
from hailhydro.flux_init import *

def initialize_overflow_on_circuit(circuit):

    overflow_landscape=overflow(circuit)

    return overflow_landscape

class overflow(flux,object):

    def __init__(self,*args):
        super(overflow,self).__init__(args[0])
        self.crit_pe=50.

    # compute concentration profile
    def calc_profile_concentration(self):

        self.update_transport_matrix()
        c,B_new=self.solve_absorbing_boundary()

        return c,B_new

    def solve_absorbing_boundary(self):

        # reduce transport matrix by cutting row,col corresponding to absorbing boundary
        B_new=self.B_eff[self.idx_eff,:]
        B_new=B_new[:,self.idx_eff]
        S=self.circuit.nodes['solute'][self.idx_eff]
        concentration=np.zeros(self.N)
        # solving inverse flux problem for absorbing boundaries
        concentration_reduced=np.dot(np.linalg.inv(B_new),S)
        concentration[self.idx_eff]=concentration_reduced[:]

        # export solution
        for i,n in enumerate(self.circuit.list_graph_nodes):
             self.circuit.G.nodes[n]['concentrations']=concentration[i]
        self.circuit.nodes['concentration']=concentration[:]

        return concentration_reduced ,B_new

    def update_transport_matrix(self):

        A=self.calc_diff_flux( self.circuit.edges['radius_sq'] )
        Q=self.calc_flow( self.circuit.edges['conductivity'], self.circuit.nodes['source'] )
        V=self.calc_velocity_from_flowrate( Q, self.circuit.edges['radius_sq'] )
        self.circuit.edges['peclet']=self.calc_peclet(V)
        self.circuit.edges['flow_rate']=Q

        x,z,e_up_sinh_x,e_down_sinh_x,coth_x,idx_pack=self.compute_flux_pars()

        f1= np.multiply(z,A)
        f2= np.multiply(A,np.multiply(x,coth_x))*0.5

        f3,f4=self.calc_absorption_jacobian_coefficients_11(0.5*np.multiply(A,x),[e_up_sinh_x,e_down_sinh_x])

        self.B_eff=np.zeros((self.N,self.N))

        for i,n in enumerate(self.circuit.list_graph_nodes):

            self.B_eff[i,i]= np.sum(  np.add( np.multiply(self.B[i,:],f1),np.multiply(np.absolute(self.B[i,:]),f2))  )
            self.B_eff[i,self.dict_in[n]]= -f3[self.dict_node_in[n]]
            self.B_eff[i,self.dict_out[n]]= -f4[self.dict_node_out[n]]

    def compute_flux_pars(self):

        x=np.sqrt( np.add( np.power(self.circuit.edges['peclet'],2),self.circuit.edges['absorption'] ) )
        z=self.circuit.edges['peclet']*0.5

        e_up_sinh_x=np.zeros(self.M)
        e_down_sinh_x=np.zeros(self.M)
        coth_x=np.zeros(self.M)

        # establish the use of converging limit expressions to prevent overflow error
        idx_lower=np.where(np.absolute(self.circuit.edges['peclet'])<self.crit_pe)[0]
        idx_over_pos=np.where((np.absolute(self.circuit.edges['peclet'])>=self.crit_pe) & (self.circuit.edges['peclet'] > 0))[0]
        idx_over_neg=np.where((np.absolute(self.circuit.edges['peclet'])>=self.crit_pe) & (self.circuit.edges['peclet'] < 0))[0]
        idx_pack=[list(idx_lower),list(idx_over_pos)+list(idx_over_neg)]

        # subcritical pe
        e_up=np.exp(z[idx_lower])
        e_down=np.exp(-z[idx_lower])
        e_up_sinh_x[idx_lower]=e_up/np.sinh(x[idx_lower]*0.5)
        e_down_sinh_x[idx_lower]=e_down/np.sinh(x[idx_lower]*0.5)
        coth_x[idx_lower]=np.reciprocal(np.tanh(x[idx_lower]*0.5))

        # overcriticial, pe positive
        e_up_sinh_x[idx_over_pos]=2.
        e_down_sinh_x[idx_over_pos]=0.
        coth_x[idx_over_pos]=1.

        # overcriticial, pe negative
        e_up_sinh_x[idx_over_neg]=0.
        e_down_sinh_x[idx_over_neg]=2.
        coth_x[idx_over_neg]=1.

        return [x,z,e_up_sinh_x,e_down_sinh_x,coth_x,idx_pack]

    # calc link absorption
    def calc_absorption(self):

        # calc coefficients
        c_a,c_b=self.get_concentrations_from_edges()
        x,z,e_up_sinh_x,e_down_sinh_x,coth_x,idx_pack=self.compute_flux_pars()

        f1_up,f1_down=self.calc_absorption_jacobian_coefficients_11(0.5*x,[e_up_sinh_x,e_down_sinh_x])
        x_coth_x=np.multiply(x,coth_x)*0.5
        F1= np.add( np.subtract(x_coth_x  , f1_up), z)
        F2= np.subtract( np.subtract(x_coth_x , f1_down), z)

        # calc edgewise absorption
        phi=np.add(np.multiply(c_a,F1) ,np.multiply( c_b,F2 ))
        A=self.calc_diff_flux( self.circuit.edges['radius_sq'] )

        PHI=np.multiply( A, phi )
        self.circuit.edges['uptake']=PHI[:]

        return PHI

    def get_concentrations_from_edges(self):

        # set containers
        c_a,c_b=np.ones(self.M),np.ones(self.M)

        for j,e in enumerate(self.circuit.list_graph_edges):

            a,b=self.dict_edges[e]
            c_a[j]=self.circuit.G.nodes[a]['concentrations']
            c_b[j]=self.circuit.G.nodes[b]['concentrations']

        return c_a,c_b

    # calc absorption jacobian and subcomponents
    def calc_absorption_jacobian(self):

        # calc coefficients
        A=self.calc_diff_flux(  self.circuit.edges['radius_sq'])
        alphas,omegas=self.get_alpha_omega_from_edges()
        c_a,c_b=self.get_concentrations_from_edges()
        # c_n=self.get_concentrations_from_nodes()
        c_n=self.circuit.nodes['concentration']

        pars=self.compute_flux_pars()
        F1,F2=self.calc_absorption_jacobian_coefficients_1(*pars)
        F3,F4=self.calc_absorption_jacobian_coefficients_2(*pars)
        qa,qb,q1,q2=self.calc_absorption_jacobian_coefficients_11(A,[c_a,c_b,F1,F2])
        phi=np.add( np.multiply(c_a,F1) ,np.multiply(c_b,F2 ) )

        # calc jacobian components
        J_PE,J_Q= self.calc_flux_jacobian()
        J_A=self.calc_cross_section_jacobian()
        J_C=self.calc_concentration_jacobian( J_PE,c_n)

        J_phi= np.zeros((self.M,self.M))
        for j,e in enumerate(self.circuit.list_graph_edges):

            J_phi[j,:]=np.add(J_phi[j,:],np.multiply( J_A[j,:], phi))

            J_phi[j,:]=np.add(J_phi[j,:], np.multiply( J_C[j,alphas], q1 ))
            J_phi[j,:]=np.add(J_phi[j,:], np.multiply( J_C[j,omegas], q2 ))

            J_phi[j,:]=np.add(J_phi[j,:],np.multiply( J_PE[j,:], np.multiply(qa,F3)))
            J_phi[j,:]=np.add(J_phi[j,:],np.multiply( J_PE[j,:], np.multiply(qb,F4)))

        return J_phi

    def get_alpha_omega_from_edges(self):

        alphas,omegas=[],[]
        for j,e in enumerate(self.circuit.list_graph_edges):

            a,b=self.dict_edges[e]
            alphas.append(a)
            omegas.append(b)

        return alphas, omegas

    def calc_absorption_jacobian_coefficients_1(self,*args):

        x,z,e_up_sinh_x,e_down_sinh_x,coth_x,idx_pack=args

        f1_up,f1_down=self.calc_absorption_jacobian_coefficients_11(0.5*x,[e_up_sinh_x,e_down_sinh_x])
        F1=np.add( np.subtract( 0.5*np.multiply(x,coth_x) ,  f1_up ), z )
        F2=np.subtract( np.subtract( 0.5*np.multiply(x,coth_x) , f1_down ), z)

        return F1, F2

    def calc_absorption_jacobian_coefficients_11(self,x,pars):

        coeff=[np.multiply(x,y) for y in pars]

        return coeff

    def calc_absorption_jacobian_coefficients_2(self, *args):

        x,z,e_up_sinh_x,e_down_sinh_x,coth_x,idx_pack=args
        F3=np.zeros(self.M)
        F4=np.zeros(self.M)
        idx_lower=idx_pack[0]
        idx_over=idx_pack[1]

        # subcritical
        sinh_x=np.sinh(x[idx_lower]*0.5)
        cosh_x=np.cosh(x[idx_lower]*0.5)
        e_up=np.exp(z[idx_lower])
        e_down=np.exp(-z[idx_lower])

        f2_up=np.subtract( np.multiply( np.divide(2.*z[idx_lower],x[idx_lower]), np.subtract( cosh_x, e_up )), np.divide(z[idx_lower],sinh_x))
        f3_up=np.add( np.multiply( e_up, np.subtract( np.multiply(coth_x[idx_lower],z[idx_lower]), 0.5*x[idx_lower] ) ) , sinh_x)
        f2_down=np.subtract(np.multiply( np.divide(2.*z[idx_lower],x[idx_lower]), np.subtract( cosh_x, e_down )), np.divide(z[idx_lower],sinh_x))
        f3_down=np.subtract( np.multiply( e_down, np.add( np.multiply(coth_x[idx_lower],z[idx_lower]), 0.5*x[idx_lower]  ) ), sinh_x )

        F3[idx_lower]= 0.5*np.divide( np.add(f2_up, f3_up) , sinh_x)
        F4[idx_lower]= 0.5*np.divide( np.add(f2_down, f3_down) , sinh_x)

        # overcritical
        f2_up= np.multiply( np.divide(2.*z[idx_over],x[idx_over]), np.subtract( coth_x[idx_over], 2. ))
        f3_up=np.add( 2.* np.subtract( np.multiply(coth_x[idx_over],z[idx_over]), 0.5*x[idx_over] )  , 1.)

        f2_down=np.multiply( np.divide(2.*z[idx_over],x[idx_over]), coth_x[idx_over] )
        # f3_down=np.zeros(len(idx_over))

        F3[idx_over]= 0.5* np.add(f2_up, f3_up)
        F4[idx_over]= 0.5* f2_down

        return F3,F4

    # calc flux jacobian and subcomponents
    def calc_flux_jacobian(self):

        # init containers
        I=np.identity(self.M)
        J_PE, J_Q= np.zeros((self.M,self.M)),np.zeros((self.M,self.M))

        # set coefficients
        f1= 2.*np.divide(self.circuit.edges['peclet'],self.circuit.edges['radius'])
        f2= 4.* np.divide(self.circuit.edges['flow_rate'],self.circuit.edges['radius'])

        INV=lina.pinv(np.dot(self.B,np.dot(np.diag(self.circuit.edges['conductivity']),self.BT)))
        D=np.dot(np.dot(self.BT,INV),self.B)

        # calc jacobian
        for i,c in enumerate(self.circuit.edges['conductivity']):

            sq_kernel=self.circuit.edges['radius_sq']/self.circuit.edges['radius_sq'][i]
            l_kernel=np.divide(self.circuit.edges['length'][i],self.circuit.edges['length'])

            J_PE[i,:]= f1[i] * np.subtract( I[i,:], 2.* c * np.multiply( D[:,i],sq_kernel) )
            J_Q[i,:]= f2[i] * np.subtract( I[i,:], c*np.multiply( D[:,i], np.multiply( l_kernel , np.power( sq_kernel , 2 ) ) ) )

        return J_PE,J_Q

    # calc cross section jacobian and subcomponents
    def calc_cross_section_jacobian(self):

        J_A=2.*np.pi*np.diag( self.circuit.edges['radius'])*self.ref_vars

        return J_A

    # calc concentraion jacobian and subcomponents
    def calc_concentration_jacobian(self, J_PE,c ):

        # set coefficients
        dict_coeff=self.calc_concentration_jacobian_coefficients( J_PE,c )
        inv_B, c = self.calc_inv_B(c)

        J_C=np.zeros((self.M,self.N))
        J_diag= np.diag(self.calc_cross_section_jacobian())

        J_A=np.zeros(self.M)
        for j,e in enumerate(self.circuit.list_graph_edges):

            J_A[j-1]=0.
            J_A[j]=J_diag[j]
            J_PE_j=J_PE[j,:]
            pars=[  [J_diag[j],J_PE_j,J_A,j] for i in range(self.N) ]

            JB_eff=np.diag( list( map(self.calc_incidence_jacobian_diag, dict_coeff['flux_sum_1'], dict_coeff['flux_sum_2'],pars) ) )
            self.calc_incidence_jacobian_dev(JB_eff,dict_coeff,pars[0])
            self.evaluate_jacobian(j,J_C,JB_eff,inv_B,c)

        return J_C

    def calc_concentration_jacobian_coefficients( self, J_PE, c  ):

        dict_coeff={}
        A=self.calc_diff_flux(  self.circuit.edges['radius_sq'] )
        x,z,e_up_sinh_x,e_down_sinh_x,coth_x,idx_pack=self.compute_flux_pars()

        f1= np.multiply(z,A)
        f2= np.multiply(np.multiply(x,coth_x),A)*0.5
        f3,f4=self.calc_absorption_jacobian_coefficients_11(0.5*np.multiply(A,x),[e_up_sinh_x,e_down_sinh_x])

        j_coth_x=np.zeros(self.M)
        idx_lower=idx_pack[0]
        j_coth_x[idx_lower]=np.power(np.divide(coth_x[idx_lower],np.cosh(x[idx_lower])),2)

        f2= np.multiply(x,coth_x)*0.5
        f4=np.subtract( np.multiply( np.divide(z,x), coth_x) ,  np.multiply( z,j_coth_x )*0.5 )
        f_up,f_down=self.calc_absorption_jacobian_coefficients_11(0.5*x,[e_up_sinh_x,e_down_sinh_x])
        f5,f6=self.calc_absorption_jacobian_coefficients_11(A,[e_up_sinh_x,e_down_sinh_x])

        J_f_up=-np.multiply( f5, np.subtract( np.add( np.divide(z,x), x*0.25 ), np.multiply(z,coth_x)*0.5 ))
        J_f_down=-np.multiply( f6, np.subtract( np.subtract( np.divide(z,x), x*0.25 ), np.multiply(z,coth_x)*0.5 ))

        dict_coeff['f_up']=f_up
        dict_coeff['f_down']=f_down
        dict_coeff['J_f_up']=J_f_up
        dict_coeff['J_f_down']=J_f_down

        flux_sum_1=np.array([ np.add( np.multiply(self.B[i,:],z), np.multiply(np.absolute(self.B[i,:]),f2)) for i,n in enumerate(self.circuit.list_graph_nodes)])
        flux_sum_2=np.array([ np.multiply(A, np.add( self.B[i,:]*0.5, np.multiply(np.absolute(self.B[i,:]),f4) ) ) for i,n in enumerate(self.circuit.list_graph_nodes)])
        dict_coeff['flux_sum_1']=flux_sum_1
        dict_coeff['flux_sum_2']=flux_sum_2

        return dict_coeff

    def calc_inv_B(self ,c):

        B_new=self.B_eff[self.idx_eff,:]
        B_new=B_new[:,self.idx_eff]
        c=c[self.idx_eff]
        inv_B=np.linalg.inv(B_new)

        return inv_B, c

    def calc_incidence_jacobian_diag(self, flux_sum_1,flux_sum_2,pars ):

        J_diag,J_PE_j,J_A,j=pars
        JB_eff=  J_diag * flux_sum_1[j] + np.sum( np.multiply( J_PE_j, flux_sum_2  ))

        return JB_eff

    def calc_incidence_jacobian_dev(self,JB_eff,dict_coeff,pars):

        J_diag,J_PE_j,J_A,j=pars
        for i,n in enumerate(self.circuit.list_graph_nodes):

            JB_eff[i,self.dict_out[n]]= np.subtract( np.multiply( J_PE_j[self.dict_node_out[n]] , dict_coeff['J_f_down'][self.dict_node_out[n]] ), np.multiply( J_A[self.dict_node_out[n]], dict_coeff['f_down'][self.dict_node_out[n]] ))

            JB_eff[i,self.dict_in[n]]=  np.subtract( np.multiply( J_PE_j[self.dict_node_in[n]] , dict_coeff['J_f_up'][self.dict_node_in[n]] ) , np.multiply( J_A[self.dict_node_in[n]], dict_coeff['f_up'][self.dict_node_in[n]] ))

    def evaluate_jacobian(self,j,J_C,JB_eff,inv_B,c):

        JB_new=JB_eff[self.idx_eff,:]
        JB_new=JB_new[:,self.idx_eff]
        J_C[j,self.idx_eff]=-np.dot(inv_B, np.dot( JB_new, c ))



# @Author:  Felix Kramer
# @Date:   2021-06-03T11:02:33+02:00
# @Email:  kramer@mpi-cbg.de
# @Project: go-with-the-flow
# @Last modified by:    Felix Kramer
# @Last modified time: 2021-09-04T23:31:01+02:00
# @License: MIT

import numpy as np
import networkx as nx
import kirchhoff.circuit_flow as kfc

# take an initiliazed circuit and start computing flows
def initialize_flow_on_circuit(circuit):

    flow_landscape=flow(circuit)

    return flow_landscape


class flow():

    def __init__(self,circuit):

        self.circuit=circuit
        self.B,self.BT=self.circuit.get_incidence_matrices()

    def find_roots(self,G):

        roots=[n for n in self.circuit.list_graph_nodes if G.nodes[n]['source']>0]

        return roots

    def find_sinks(self,G):

        sinks=[n for n in self.circuit.list_graph_nodes if G.nodes[n]['source']<0]

        return sinks

    def alpha_omega(self,G,j):

        labels=nx.get_edge_attributes(G,'label')
        for e,label in labels.items():
            if label==j:
                alpha=e[1]
                omega=e[0]

        return alpha,omega

    def calc_pressure(self,*args):

        conduct,source=args

        OP=np.dot(self.B,np.dot(np.diag(conduct),self.BT))
        P,RES,RG,si=np.linalg.lstsq(OP,source,rcond=None)
        dP=np.dot(self.BT,P)

        return dP, P

    def calc_flow_from_pressure(self,*args):

        conduct,dP=args
        Q=np.dot(np.diag(conduct),dP)

        return Q

    def calc_flow(self,*args):

        conduct,source=args

        dP,P=self.calc_pressure(*args)
        Q=np.dot(np.diag(conduct),dP)

        return Q

    def calc_sq_flow(self,*args):

        dP,P=self.calc_pressure(*args)
        Q=self.calc_flow_from_pressure(args[0],dP)

        p_sq=np.multiply(dP,dP)
        q_sq=np.multiply(Q,Q)

        return p_sq, q_sq

    def calc_cross_section_from_conductivity(self,*args):

        conductivity,conductance=args
        R_sq=np.sqrt(conductivity/conductance)

        return R_sq

    def calc_conductivity_from_cross_section(self,*args):

        R_sq,conductance=args
        conductivity=np.power(R_sq,2)*conductance

        return conductivity

    def calc_configuration_flow(self):

        k= self.circuit.edges['conductivity']
        src= self.circuit.nodes['source']

        dP,P=self.calc_pressure(k,src)
        Q=np.dot(np.diag(k),dP)

        return Q



# @Author:  Felix Kramer
# @Date:   2021-06-03T11:02:57+02:00
# @Email:  kramer@mpi-cbg.de
# @Project: go-with-the-flow
# @Last modified by:    Felix Kramer
# @Last modified time: 2021-08-30T23:39:30+02:00
# @License: MIT

import sys
import networkx as nx
import numpy as np
import scipy.linalg as lina
import random as rd
from hailhydro.flux_init import *
from hailhydro.flow_random import *
from hailhydro.flux_overflow import *

class flux_random(overflow ,flow_reroute,object):

    def __init__(self,*args):
        super(flux_random,self).__init__(args[0])

    def calc_flow(self,*args):

        graph_matrices=self.get_broken_links_asarray(*args)
        flow_observables=list(map(self.calc_flows_mapping,graph_matrices))

        return flow_observables

    def calc_transport_observables(self,*args):

        # calc ensemble averages
        idx,conduct,flow_observables=args

        graph_matrices=self.get_broken_links_asarray(idx,conduct)
        R_powers=calc_random_radii(idx,conduct)
        p_sq=np.power([fo[2] for fo in flow_observables],2)

        PHI=list(map( self.calc_noisy_absorption , R_powers[1], flow_observables ) )
        SHEAR=np.multiply(dV_sq,R_powers[1])

        avg_shear_sq=np.sum(np.multiply(dV_sq,R_sq),axis=0)/float(self.num_iteration)
        avg_PHI= np.mean(PHI,axis=0)

        return shear,phi

    def calc_noisy_absorption(self, R_sq, flow_observables):

        self.update_transport_matrix(R_sq ,flow_observables)

        c,B_new=self.solve_absorbing_boundary()

        return self.calc_absorption(R_sq)

    def update_transport_matrix(self,R_sq,flow_observables):

        # set peclet number and internal flow state
        self.circuit.edge['flow_rate']=flow_observables[0]
        ref_var=self.circuit.scale['length']/self.circuit.scale['diffusion']

        V=self.calc_velocity_from_flowrate(self.circuit.edge['flow_rate'],R_sq)
        self.circuit.edge['peclet']=self.calc_peclet( V,ref_var )
        A=self.calc_diff_flux( R_sq,1./ref_var )

        x,z,e_up_sinh_x,e_down_sinh_x,coth_x,idx_pack = self.compute_flux_pars()

        f1= np.multiply(z,A)
        f2= np.multiply(A,np.multiply(x,coth_x))*0.5

        f3= np.multiply(np.multiply(A,x),e_up_sinh_x)*0.5
        f4= np.multiply(np.multiply(A,x),e_down_sinh_x)*0.5

        # set up concentration_matrix
        self.B_eff=np.zeros((self.N,self.N))

        for i,n in enumerate(self.circuit.list_graph_nodes):

            self.B_eff[i,i]= np.sum(  np.add( np.multiply(self.B[i,:],f1),np.multiply(np.absolute(self.B[i,:]),f2))  )
            self.B_eff[i,self.dict_in[n]]= -f3[self.dict_node_in[n]]
            self.B_eff[i,self.dict_out[n]]= -f4[self.dict_node_out[n]]



# @Author:  Felix Kramer
# @Date:   2021-06-03T11:02:33+02:00
# @Email:  kramer@mpi-cbg.de
# @Project: go-with-the-flow
# @Last modified by:    Felix Kramer
# @Last modified time: 2021-08-30T23:39:31+02:00
# @License: MIT

import numpy as np
import scipy.linalg as lina
import sys

import random as rd
import networkx as nx
from hailhydro.flow_init import *

def initialize_random_flow_on_circuit(circuit,flow_setting={'mode':'default', 'noise':0.}):

    flow_landscape=flow_random(circuit,flow_setting)

    return flow_landscape

def initialize_rerouting_flow_on_circuit(circuit,flow_setting={'p_broken':0, 'num_iter':100}):

    flow_landscape=flow_reroute(circuit,flow_setting)

    return flow_landscape

class flow_random(flow,object):

    def __init__(self,circuit,flow_setting):

        super(flow_random,self).__init__(circuit)

        if flow_setting['mode']=='default':
            self.set_effective_source_matrix(flow_setting['noise'])

        if flow_setting['mode']=='root':
            self.set_multi_source_matrix(flow_setting['mu_sq'],flow_setting['var'])

    # setup_random_fluctuations
    def set_root_source_matrix(self,*args):

        mean, variance= args
        N=len(self.circuit.list_graph_nodes)

        self.matrix_mu=np.identity()
        for n in range(N):
            for m in range(N):
                h=0.
                if n==0 and m==0:
                    h+=(N-1)
                elif n==m and n!=0:
                    h+=1.
                elif n==0 and m!=0:
                    h-=1.
                elif m==0 and n!=0:
                    h-=1.

                self.matrix_mu[n,m]=h

        self.matrix_mu=self.matrix_mu*mean

        self.matrix_var=np.identity(N)
        for n in range(N):

            for m in range(N):
                h=0.
                if n==0 and m==0:
                    h+=(1-N)*(1-N)
                elif n!=0 and m!=0:
                    h+=1.
                elif n==0 and m!=0:
                    h+=(1-N)
                elif m==0 and n!=0:
                    h+=(1-N)

                self.matrix_var[n,m]=h

        self.matrix_var=self.matrix_var*variance

    # setup_random_fluctuations_multisink
    def set_effective_source_matrix(self, noise ):

        self.noise=noise
        num_n=len(self.circuit.list_graph_nodes)
        x=np.where( self.circuit.nodes['source'] > 0)[0]
        idx=np.where( self.circuit.nodes['source'] < 0)[0]
        N=len(idx)
        M=len(x)

        U=np.zeros((num_n,num_n))
        V=np.zeros((num_n,num_n))

        m_sq=float(M*M)
        NM=num_n*num_n/float(m_sq)
        Nm=(N/m_sq)+2./M

        for i in range(num_n):
            for j in range(num_n)[i:]:
                delta=0.
                sum_delta=0.
                sum_delta_sq=0.

                if i==j:
                    delta=1.

                if (i in x):
                    sum_delta=1.

                if (j in x):
                    sum_delta=1.

                if (i in x) and (j in x):
                    sum_delta_sq=1.
                    sum_delta=2.

                U[i,j]= ( m_sq - num_n*sum_delta + NM*sum_delta_sq )
                V[i,j]= ( ( Nm + delta )*sum_delta_sq - (1.+M*delta)*sum_delta + m_sq*delta)

                U[j,i]=U[i,j]
                V[j,i]=V[i,j]

        self.Z = np.add(U,np.multiply(self.noise,V))

     # calc_sq_flow
    def calc_sq_flow_effective(self,conduct):

        OP=np.dot(self.B,np.dot(np.diag(conduct),self.BT))
        inverse=lina.pinv(OP)
        D=np.dot(self.BT,inverse)
        DT=np.transpose(D)

        A=np.dot(np.dot(D,self.Z),DT)
        dV_sq=np.diag(A)
        F_sq=np.multiply(np.multiply(conduct,conduct),dV_sq)

        return dV_sq,F_sq
    # calc_sq_flow_random
    def calc_sq_flow_root(self,conduct):

        OP=np.dot(np.dot(self.B,conduct),self.BT)
        inverse=lina.pinv(OP)
        D=np.dot(self.BT,inverse)
        DT=np.transpose(D)

        var_matrix=np.dot(np.dot(D,self.matrix_mu),DT)
        mean_matrix=np.dot(np.dot(D,self.matrix_var),DT)

        var_flow=np.diag(var_matrix)
        mean_flow=np.diag(mean_matrix)

        dV_sq= np.add(var_flow , mean_flow)
        F_sq=np.multiply(np.multiply(conduct,conduct),dV_sq)

        return dV_sq,F_sq

class flow_reroute(flow,object):

    def __init__(self,circuit,flow_setting):

        super( flow_reroute,self).__init__(circuit)

        self.num_iteration=flow_setting['num_iter']
        self.percentage_broken=flow_setting['p_broken']
        self.initialize_broken_link()

    def initialize_broken_link(self):

        broken_sets=[]
        self.num_sets=50000
        self.AUX=nx.Graph(self.circuit.G)
        for i in range(self.num_sets):
            cond,idx=self.generate_coherent_closure()
            if cond:
                broken_sets.append(idx)

        self.broken_sets=broken_sets
        if len(self.broken_sets)==0:
            sys.exit('nothing broken here... srsly check initialize_broken_link() though')

    def generate_coherent_closure_deterministic(self,H,x):

        idx=rd.sample(range(len(self.circuit.list_graph_edges)),x)
        for e in idx:
            H.remove_edge(*self.circuit.list_graph_edges[e])
        cond=nx.is_connected(H)

        for e in idx:
            H.add_edge(*self.circuit.list_graph_edges[e])

        return cond,idx

    def generate_coherent_closure(self):

        prob=np.random.sample(len(self.circuit.list_graph_edges))
        idx=np.where(   prob <= self.percentage_broken )[0]

        for e in idx:
            self.AUX.remove_edge(*self.circuit.list_graph_edges[e])
        cond=nx.is_connected(self.AUX)

        for e in idx:
            self.AUX.add_edge(*self.circuit.list_graph_edges[e])

        return cond,idx

    def break_links(self,*args):

        idx,conduct=args

        C_aux=np.array(conduct)
        C_aux[idx]=np.power(10.,-20)

        return C_aux

    def get_sets(self):

        idx=rd.choices(self.broken_sets,k=self.num_iteration)

        return idx

    def calc_random_radii(self,*args):

        graph_matrices=self.get_broken_links_asarray(*args)

        R,R_sq,R_cb=[],[],[]
        for gm in graph_matrices:

            kernel=gm/self.circuit.scales['conductance']

            R.append(np.power(kernel,0.25))
            R_sq.append(np.sqrt(kernel))
            R_cb.append(np.power(kernel,0.75))

        return [R,R_sq, R_cb]

    def calc_sq_flow(self,*args):

        # block p percent of the edges per realization
        graph_matrices=self.get_broken_links_asarray(*args)
        flow_observables=list(map(self.calc_flows_mapping,graph_matrices))

        # calc ensemble averages
        q_sq=np.power([fo[0] for fo in flow_observables],2)
        p_sq=np.power([fo[2] for fo in flow_observables],2)

        return p_sq,q_sq

    def calc_flows_mapping(self,graph_matrices):

        C_aux=graph_matrices

        dP, P=self.calc_pressure(C_aux,self.circuit.nodes['source'])
        Q=self.calc_flow_from_pressure(C_aux,dP)

        return [Q,P,dP]

    def calc_sq_flow_avg(self,conduct):

        idx=rd.choices(self.broken_sets,k=self.num_iteration)

        p_sq,q_sq=self.calc_sq_flow(idx,conduct)
        R,R_sq, R_cb=self.calc_random_radii(idx,conduct)

        avg_diss=np.sum(np.multiply(p_sq,R_cb),axis=0)/float(self.num_iteration)
        avg_R=np.mean(R,axis=0)
        avg_dP_sq=np.mean(p_sq,axis=0)
        avg_F_sq= np.mean(q_sq,axis=0)

        return [avg_dP_sq,avg_F_sq,avg_R,avg_diss]

    def get_broken_links_asarray(self, *args):

        idx,conduct=args

        graph_matrices=[self.break_links(i,conduct) for i in idx]

        return graph_matrices



# @Author:  Felix Kramer <kramer>
# @Date:   2021-05-08T20:18:09+02:00
# @Email:  kramer@mpi-cbg.de
# @Project: go-with-the-flow
# @Last modified by:   kramer
# @Last modified time: 2021-05-09T00:48:20+02:00
# @License: MIT



# @Author:  Felix Kramer
# @Date:   2021-05-23T23:25:28+02:00
# @Email:  kramer@mpi-cbg.de
# @Project: go-with-the-flow
# @Last modified by:    Felix Kramer
# @Last modified time: 2021-09-05T00:06:13+02:00
# @License: MIT



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hailhydro", # Replace with your own username
    version="0.0.5",
    author="felixk1990",
    author_email="felixuwekramer@protonmail.com",
    description="Collection of tools for computation of flows and and fluxes on Kirchhoff circuits. ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/felixk1990/kirchhoff-hydro",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



# @Author:  Felix Kramer
# @Date:   2021-06-03T11:02:57+02:00
# @Email:  kramer@mpi-cbg.de
# @Project: go-with-the-flow
# @Last modified by:    Felix Kramer
# @Last modified time: 2021-08-30T23:39:32+02:00
# @License: MIT

import networkx as nx
import numpy as np
import scipy.linalg as lina
from hailhydro.flow_init import *
from hailhydro.flow_random import *

def initialize_flux_on_circuit(circuit):

    flux_landscape=flux(circuit)

    return flux_landscape

class flux(flow, object):

    def __init__(self,*args):
        super(flux,self).__init__(args[0])

        # incidence correlation
        self.dict_in={}
        self.dict_out={}
        self.dict_edges={}

        # incidence indices
        self.dict_node_out={}
        self.dict_node_in={}

        self.initialize()

    def initialize(self):

        self.ref_vars=self.circuit.scales['diffusion']/self.circuit.scales['length']
        self.N=len(self.circuit.list_graph_nodes)
        self.M=len(self.circuit.list_graph_edges)
        self.circuit.nodes['concentration']=np.zeros(self.N)

        self.roots=self.find_roots(self.circuit.G)
        self.sinks=self.find_sinks(self.circuit.G)
        self.nodes_sinks=[self.circuit.G.nodes[sink]['label'] for sink in self.sinks]
        self.nodes_roots=[self.circuit.G.nodes[source]['label'] for source in self.roots]

        self.idx_eff=[i for i in range(self.N) if i not in self.nodes_sinks]

        for i,n in  enumerate(self.circuit.list_graph_nodes):
            self.dict_in[n]=[]
            self.dict_out[n]=[]
            self.dict_node_out[n]=np.where(self.B[i,:]>0)[0]
            self.dict_node_in[n]=np.where(self.B[i,:]<0)[0]

        for j,e in  enumerate(self.circuit.list_graph_edges):

            alpha=e[1]
            omega=e[0]
            if self.B[alpha,j] > 0.:

                self.dict_edges[e]=[alpha,omega]
                self.dict_in[omega].append(alpha)
                self.dict_out[alpha].append(omega)

            elif self.B[alpha,j] < 0.:

                self.dict_edges[e]=[omega,alpha]
                self.dict_in[alpha].append(omega)
                self.dict_out[omega].append(alpha)

            else:
                print('and I say...whats going on? I say heyayayayayaaaaaaa...')

    def calc_diff_flux(self,R_sq):

        A=np.pi*R_sq*self.ref_vars

        return A

    def calc_velocity_from_flowrate(self,*args):

        Q,R_sq=args
        V=np.divide(Q,R_sq*np.pi)

        return V

    def calc_peclet(self,V):

        PE=V/self.ref_vars

        return PE
